@@ -1,5 +1,5 @@
 /* copy.c -- core functions for copying files and directories
-   Copyright (C) 89, 90, 91, 1995-2002 Free Software Foundation.
+   Copyright (C) 89, 90, 91, 1995-2003 Free Software Foundation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -83,13 +83,13 @@ struct F_triple
 int euidaccess ();
 int yesno ();
 
-static int copy_internal PARAMS ((const char *src_path, const char *dst_path,
-				  int new_dst, dev_t device,
-				  struct dir_list *ancestors,
-				  const struct cp_options *x,
-				  int command_line_arg,
-				  int *copy_into_self,
-				  int *rename_succeeded));
+static int copy_internal (const char *src_path, const char *dst_path,
+			  int new_dst, dev_t device,
+			  struct dir_list *ancestors,
+			  const struct cp_options *x,
+			  int command_line_arg,
+			  int *copy_into_self,
+			  int *rename_succeeded);
 
 /* Pointers to the file names:  they're used in the diagnostic that is issued
    when we detect the user is trying to copy a directory into itself.  */
@@ -1110,14 +1110,32 @@ copy_internal (const char *src_path, const char *dst_path,
 	  goto un_backup;
 	}
 
-      if (link (earlier_file, dst_path))
-	{
-	  error (0, errno, _("cannot create hard link %s to %s"),
-		 quote_n (0, dst_path), quote_n (1, earlier_file));
-	  goto un_backup;
-	}
+      {
+	int link_failed;
 
-      return 0;
+	link_failed = link (earlier_file, dst_path);
+
+	/* If the link failed because of an existing destination,
+	   remove that file and then call link again.  */
+	if (link_failed && errno == EEXIST)
+	  {
+	    if (unlink (dst_path))
+	      {
+		error (0, errno, _("cannot remove %s"), quote (dst_path));
+		goto un_backup;
+	      }
+	    link_failed = link (earlier_file, dst_path);
+	  }
+
+	if (link_failed)
+	  {
+	    error (0, errno, _("cannot create hard link %s to %s"),
+		   quote_n (0, dst_path), quote_n (1, earlier_file));
+	    goto un_backup;
+	  }
+
+	return 0;
+      }
     }
 
   if (x->move_mode)
@@ -1538,11 +1556,14 @@ copy_internal (const char *src_path, const char *dst_path,
 
 un_backup:
 
-  /* We didn't create the destination.
-     Remove the entry associating the source dev/ino with the
+  /* We have failed to create the destination file.
+     If we've just added a dev/ino entry via the remember_copied
+     call above (i.e., unless we've just failed to create a hard link),
+     remove the entry associating the source dev/ino with the
      destination file name, so we don't try to `preserve' a link
      to a file we didn't create.  */
-  forget_created (src_sb.st_ino, src_sb.st_dev);
+  if (earlier_file == NULL)
+    forget_created (src_sb.st_ino, src_sb.st_dev);
 
   if (dst_backup)
     {
